# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect

from .base import BaseConstraint


class FunctionConstraint(BaseConstraint):
    def __init__(self, func, description=None):
        super().__init__(description=description)
        # Check if the function has exactly one parameter
        sig = inspect.signature(func)

        if len(sig.parameters) != 1:
            raise ValueError(f"The function {func.__qualname__} must have exactly one parameter.")

        self.func = func
        self.description = description or func.__name__
        
    def __str__(self):
        return self.description

    def is_satisfied_by(self, val):
        result = self.func(val)

        if not isinstance(result, bool):
            raise TypeError("The function must return a boolean value.")
        return result


class AndConstraint(BaseConstraint):
    def __init__(self, *constraints, description=None):
        super().__init__(description=description)

        self.constraints = constraints
        self.description = " and ".join(str(c) for c in constraints)

    def is_satisfied_by(self, val):
        return all(c.is_satisfied_by(val) for c in self.constraints)


class OrConstraint(BaseConstraint):
    def __init__(self, *constraints, description=None):
        super().__init__(description=description)
        
        self.constraints = constraints
        self.description = " or ".join(str(c) for c in constraints)

    def is_satisfied_by(self, val):
        return any(c.is_satisfied_by(val) for c in self.constraints)
    
    
class NotConstraint(BaseConstraint):
    def __init__(self, constraint, *, description=None):
        super().__init__(description=description)
        
        self.constraint = constraint
        self.description = f"not {self.constraint}"

    def is_satisfied_by(self, val):
        return not self.constraint.is_satisfied_by(val)


class IfElseConstraint(BaseConstraint):
    def __init__(self, condition, if_constraint, else_constraint, *, description=None):
        super().__init__(description=description)
        
        if isinstance(condition, bool):
            if description is None:
                raise ValueError("'description' must not be None when 'condition' is a function")
            self.condition = FunctionConstraint(lambda _: condition, description)
        elif isinstance(condition, BaseConstraint):
            self.condition = condition  
        else:
            raise TypeError(f"'condition' must be bool or BaseConstraint. Got {type(condition).__name__} instead")

        self.if_constraint = if_constraint
        self.else_constraint = else_constraint
        self.description = f"if {self.condition.description}, then {if_constraint}. Otherwise {else_constraint}"

    def is_satisfied_by(self, val):
        if self.condition.is_satisfied_by(val):
            return self.if_constraint.is_satisfied_by(val)
        else:
            return self.else_constraint.is_satisfied_by(val)



# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import Counter
from ms_service_profiler import Profiler, Level
from .vllm_hooker_base import VLLMHookerBase


def compare_deques(queue1, queue2):
    counter1 = Counter(queue1)
    counter2 = Counter(queue2)
    diff = counter1 - counter2
    return diff


def queue_profiler(before_queue, after_queue, queue_name):
    # 队列元素减少
    less_queue = compare_deques(before_queue, after_queue)
    rid_list = []
    for seq_group in less_queue:
        rid_list.append(seq_group.request_id)
    if len(rid_list) > 0:
        prof = Profiler(Level.INFO).domain("BatchSchedule").res(rid_list)
        prof.metric("QueueSize", len(after_queue)).metric_scope(queue_name).event("Dequeue")

    # 队列元素增加
    add_queue = compare_deques(after_queue, before_queue)
    rid_list.clear()
    for seq_group in add_queue:
        rid_list.append(seq_group.request_id)
    if len(rid_list) > 0:
        prof = Profiler(Level.INFO).domain("BatchSchedule").res(rid_list)
        prof.metric("QueueSize", len(after_queue)).metric_scope(queue_name).event("Enqueue")


class SchedulerHook(VLLMHookerBase):
    vllm_version = ("0.6.3", "0.8.4")

    def init(self):
        from vllm.core.scheduler import Scheduler
        from vllm.sequence import SequenceGroupMetadata, SequenceStatus

        def schedule_maker(ori_func):
            def schedule(this, *args, **kwargs):
                prof = Profiler(Level.INFO).domain("BatchSchedule").span_start("batchFrameworkProcessing")
                seq_group_metadata_list, scheduler_outputs, allow_async_output_proc = ori_func(this, *args, **kwargs)

                iter_size_map, is_prefill = {}, False
                for seq_group in this.running:
                    if len(seq_group.seqs) == 0:
                        continue
                    prompt_len = len(seq_group.prompt_token_ids)
                    generated_len = len(seq_group.seqs[0].get_token_ids()) - prompt_len
                    iter_size_map[seq_group.request_id] = generated_len
                    is_prefill = is_prefill or (generated_len == 0)
                prof.attr("batch_type", "Prefill" if is_prefill else "Decode")

                rid_list = []
                for metadata in seq_group_metadata_list:
                    if isinstance(metadata, SequenceGroupMetadata):
                        iter_size = iter_size_map.get(metadata.request_id, metadata.token_chunk_size)
                        data = {"rid": metadata.request_id, "iter_size": iter_size}
                        rid_list.append(data)
                prof.res(rid_list)
                prof.span_end()

                return seq_group_metadata_list, scheduler_outputs, allow_async_output_proc

            return schedule

        self.do_hook([Scheduler.schedule], schedule_maker)

        def abort_seq_group_maker(ori_func):
            def abort_seq_group(this, request_id, *args, **kwargs):
                prof = Profiler(Level.INFO).domain("BatchSchedule").res(request_id)
                prof.metric_inc("FINISHED_ABORTED", 1).event("ReqState")
                ori_func(this, request_id, *args, **kwargs)

            return abort_seq_group

        self.do_hook([Scheduler.abort_seq_group], abort_seq_group_maker)

        def allocate_and_set_running_maker(ori_func):
            def _allocate_and_set_running(this, seq_group, *args, **kwargs):
                prof = Profiler(Level.INFO).domain("BatchSchedule").res(seq_group.request_id)
                prof.metric_inc("RUNNING", 1).metric_inc("WAITING", -1).event("ReqState")
                ori_func(this, seq_group, *args, **kwargs)

            return _allocate_and_set_running

        self.do_hook([Scheduler._allocate_and_set_running], allocate_and_set_running_maker)

        def preempt_by_recompute_maker(ori_func):
            def _preempt_by_recompute(this, seq_group, *args, **kwargs):
                seqs = seq_group.get_seqs(status=SequenceStatus.RUNNING)
                if len(seqs) != 1:
                    prof = Profiler(Level.INFO).domain("BatchSchedule").res(seq_group.request_id)
                    prof.metric_inc("RUNNING", -1).metric_inc("WAITING", 1).event("ReqState")
                ori_func(this, seq_group, *args, **kwargs)

            return _preempt_by_recompute

        self.do_hook([Scheduler._preempt_by_recompute], preempt_by_recompute_maker)

        def swap_in_maker(ori_func):
            def _swap_in(this, seq_group, *args, **kwargs):
                prof = Profiler(Level.INFO).domain("BatchSchedule").res(seq_group.request_id)
                prof.metric_inc("RUNNING", 1).metric_inc("SWAPPED", -1).event("ReqState")
                ori_func(this, seq_group, *args, **kwargs)

            return _swap_in

        self.do_hook([Scheduler._swap_in], swap_in_maker)

        def swap_out_maker(ori_func):
            def _swap_out(this, seq_group, *args, **kwargs):
                if this.block_manager.can_swap_out(seq_group):
                    prof = Profiler(Level.INFO).domain("BatchSchedule").res(seq_group.request_id)
                    prof.metric_inc("RUNNING", -1).metric_inc("SWAPPED", 1).event("ReqState")
                ori_func(this, seq_group, *args, **kwargs)

            return _swap_out

        self.do_hook([Scheduler._swap_out], swap_out_maker)

        def free_finished_seq_groups_maker(ori_func):
            def free_finished_seq_groups(this, *args, **kwargs):
                before_running_queue = this.running
                for seq_group in this.running:
                    prof = Profiler(Level.INFO).domain("BatchSchedule").res(seq_group.request_id)
                    prof.metric_inc("RUNNING", -1).metric_inc("FINISHED", 1).event("ReqState")
                ori_func(this, *args, **kwargs)
                queue_profiler(before_running_queue, this.running, "running")

            return free_finished_seq_groups

        self.do_hook([Scheduler.free_finished_seq_groups], free_finished_seq_groups_maker)

        def add_seq_group_to_running_maker(ori_func):
            def _add_seq_group_to_running(this, seq_group, *args, **kwargs):
                ori_func(this, seq_group, *args, **kwargs)
                prof = Profiler(Level.INFO).domain("BatchSchedule").res([seq_group.request_id])
                prof.metric("QueueSize", len(this.running)).metric_scope("running").event("Enqueue")

            return _add_seq_group_to_running

        self.do_hook([Scheduler._add_seq_group_to_running], add_seq_group_to_running_maker)

        def schedule_priority_preemption_maker(ori_func):
            def _schedule_priority_preemption(this, budget, *args, **kwargs):
                before_waiting_queue = this.waiting
                before_running_queue = this.running
                force_preemption_count = ori_func(this, budget, *args, **kwargs)
                queue_profiler(before_waiting_queue, this.waiting, "waiting")
                queue_profiler(before_running_queue, this.running, "running")
                return force_preemption_count

            return _schedule_priority_preemption

        self.do_hook([Scheduler._schedule_priority_preemption], schedule_priority_preemption_maker)

        def schedule_default_maker(ori_func):
            def _schedule_default(this, *args, **kwargs):
                before_swapped_queue = this.swapped
                before_running_queue = this.running
                before_waiting_queue = this.waiting
                scheduler_outputs = ori_func(this, *args, **kwargs)
                queue_profiler(before_swapped_queue, this.swapped, "swapped")
                queue_profiler(before_running_queue, this.running, "running")
                queue_profiler(before_waiting_queue, this.waiting, "waiting")
                return scheduler_outputs

            return _schedule_default

        self.do_hook([Scheduler._schedule_default], schedule_default_maker)

        def schedule_chunked_prefill_maker(ori_func):
            def _schedule_chunked_prefill(this, *args, **kwargs):
                before_running_queue = this.running
                before_waiting_queue = this.waiting
                before_swapped_queue = this.swapped
                scheduler_outputs = ori_func(this, *args, **kwargs)
                queue_profiler(before_running_queue, this.running, "running")
                queue_profiler(before_waiting_queue, this.waiting, "waiting")
                queue_profiler(before_swapped_queue, this.swapped, "swapped")
                return scheduler_outputs

            return _schedule_chunked_prefill

        self.do_hook([Scheduler._schedule_chunked_prefill], schedule_chunked_prefill_maker)

        def add_seq_group_maker(ori_func):
            def add_seq_group(this, seq_group, *args, **kwargs):
                ori_func(this, seq_group, *args, **kwargs)
                prof = Profiler(Level.INFO).domain("BatchSchedule").res([seq_group.request_id])
                prof.metric("QueueSize", len(this.waiting)).metric_scope("waiting").event("Enqueue")

            return add_seq_group

        self.do_hook([Scheduler.add_seq_group], add_seq_group_maker)

        def add_seq_group_to_swapped_maker(ori_func):
            def _add_seq_group_to_swapped(this, seq_group, *args, **kwargs):
                ori_func(this, seq_group, *args, **kwargs)
                prof = Profiler(Level.INFO).domain("BatchSchedule").res([seq_group.request_id])
                prof.metric("QueueSize", len(this.swapped)).metric_scope("swapped").event("Enqueue")

            return _add_seq_group_to_swapped

        self.do_hook([Scheduler._add_seq_group_to_swapped], add_seq_group_to_swapped_maker)


class LLMEngineHook(VLLMHookerBase):
    vllm_version = ("0.6.3", "0.8.4")

    def init(self):
        from vllm.engine.llm_engine import LLMEngine

        def add_processed_request_maker(ori_func):
            def _add_processed_request(this, request_id, *args, **kwargs):
                ori_func(this, request_id, *args, **kwargs)
                Profiler(Level.INFO).domain("Request").res(request_id).metric_inc("WAITING", 1).event("ReqState")

            return _add_processed_request

        self.do_hook([getattr(LLMEngine, "_add_processed_request")], add_processed_request_maker)


batch_hookers = [SchedulerHook, LLMEngineHook]



# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import threading
from collections import defaultdict
from pathlib import Path

import numpy as np
import torch
from loguru import logger

import msserviceprofiler.modelevalstate.inference.simulate as simulate
from msserviceprofiler.modelevalstate.inference.constant import BatchStage
from msserviceprofiler.modelevalstate.inference.data_format_v1 import BatchField, RequestField
from msserviceprofiler.modelevalstate.inference.dataset import CustomLabelEncoder, preset_category_data, DataProcessor
from msserviceprofiler.modelevalstate.inference.file_reader import FileHanlder, StaticFile
from msserviceprofiler.modelevalstate.inference.simulate import ServiceField, write_file, file_log


class SimulateVllm:
    first = True
    predict_cache = {}
    req_to_output_len = defaultdict(int)
    req_to_stop_token_ids = {}
    req_id_to_max_token_by_sequence = {}
 
    @staticmethod
    def init():
        if SimulateVllm.first:
            if ServiceField.config_path.req_and_decode_file.exists():
                with open(ServiceField.config_path.req_and_decode_file, 'r') as f:
                    ServiceField.req_id_and_max_decode_length = {int(k): int(v) for k, v in json.load(f).items()}
            else:
                ServiceField.req_id_and_max_decode_length = {}
            if not Path(ServiceField.config_path.static_file_dir).exists():
                Path(ServiceField.config_path.static_file_dir).mkdir(parents=True)
            static_file = StaticFile(base_path=ServiceField.config_path.static_file_dir)
            ServiceField.fh = FileHanlder(static_file)
            ServiceField.fh.load_static_data()
            custom_encoder = CustomLabelEncoder(preset_category_data, save_dir=ServiceField.config_path.ohe_path)
            custom_encoder.fit(load=True)
            ServiceField.data_processor = DataProcessor(custom_encoder)
            SimulateVllm.first = False
            simulate.sub_thread = threading.Thread(target=write_file, args=(file_log,))
            simulate.sub_thread.start()
 
    @staticmethod
    def generate_features(model_input):
        # model_input的类是vllm_ascend.worker.model_runner.ModelInputForNPUWithSamplingMetadata
        if model_input.is_prompt is None and model_input.seq_lens is None:
            return None, None
        if model_input.is_prompt:
            batch_stage = BatchStage.PREFILL
        else:
            batch_stage = BatchStage.DECODE
        if model_input.attn_metadata.block_tables is None:
            all_need_blocks = torch.zeros((len(model_input.seq_lens), 1))
        else:
            all_need_blocks = torch.count_nonzero(model_input.attn_metadata.block_tables > -1, dim=-1)
        _total_req_input_len = []
        _all_request_ids = set({})
        _req_to_input_len = defaultdict(int)
        SimulateVllm.req_to_output_len = defaultdict(int)
        _req_to_need_blocks = defaultdict(int)
        _seq_id_to_req = {}
        for _req_id, _seq_ids in model_input.request_ids_to_seq_ids.items():
            for _id in _seq_ids:
                _seq_id_to_req[_id] = _req_id
        for _index, _seq_group_sample in enumerate(model_input.sampling_metadata.seq_groups):
            # _seq_group_sample的类是vllm.model_executor.sampling_metadata.SequenceGroupToSample
            _stop_ids = _seq_group_sample.sampling_params.stop_token_ids
            for _seq_id in _seq_group_sample.seq_ids:
                if _seq_id not in _seq_group_sample.seq_data:
                    logger.warning(f"{_seq_id} not in {_seq_group_sample.seq_data}")
                    continue
                _cur_req_id = _seq_id_to_req.get(_seq_id)
                if _stop_ids:
                    SimulateVllm.req_to_stop_token_ids[_cur_req_id] = _stop_ids
 
                _req_input_len = len(_seq_group_sample.seq_data.get(_seq_id).prompt_token_ids)
                _req_to_input_len[_cur_req_id] += _req_input_len
                _total_req_input_len.append(_req_input_len)
                _req_output_len = len(_seq_group_sample.seq_data.get(_seq_id).output_token_ids)
                SimulateVllm.req_to_output_len[_cur_req_id] += _req_output_len
                _req_need_block = all_need_blocks[_index].sum().item()
                _req_to_need_blocks[_cur_req_id] += _req_need_block
                _all_request_ids.add(_cur_req_id)
        request_field = []
        for _cur_id in _all_request_ids:
            request_field.append(RequestField(_req_to_input_len.get(_cur_id),
                                              _req_to_need_blocks.get(_cur_id),
                                              SimulateVllm.req_to_output_len.get(_cur_id)))
 
        batch_field = BatchField(batch_stage, len(model_input.seq_lens),
                                 all_need_blocks.sum().item(),
                                 sum(_total_req_input_len), max(_total_req_input_len))
 
        ServiceField.batch_field = batch_field
        ServiceField.request_field = tuple(sorted(request_field))
        return batch_field, tuple(sorted(request_field))

    @staticmethod
    def get_max_output_len(req_id):
        """
        req_id: 离线情况下，会是数字0,1,2， 在线模型会是字符串 cmpl-7d6e773db843411985fba579778e81ea-0
        """
        # 尝试直接从json文件中获取
        _max_out_len = ServiceField.req_id_and_max_decode_length.get(req_id)
        if _max_out_len is None:
            try:
                # 尝试转为int类型进行获取
                _max_out_len = ServiceField.req_id_and_max_decode_length[int(req_id)]
            except ValueError:
                pass
        # 尝试从缓存中获取
        if _max_out_len is None:
            _max_out_len = SimulateVllm.req_id_to_max_token_by_sequence.get(req_id)
        # 按照收到请求的序号，从采集收集的json文件里面获取对应序号的数据。
        if _max_out_len is None:
            _max_out_len = ServiceField.req_id_and_max_decode_length.get(
                len(SimulateVllm.req_id_to_max_token_by_sequence))
            if _max_out_len:
                SimulateVllm.req_id_to_max_token_by_sequence[req_id] = _max_out_len
        if _max_out_len is None:
            _max_out_len = ServiceField.req_id_and_max_decode_length.get(
                str(len(SimulateVllm.req_id_to_max_token_by_sequence)))
            if _max_out_len:
                SimulateVllm.req_id_to_max_token_by_sequence[req_id] = _max_out_len
        return _max_out_len
 
    @staticmethod
    def get_cur_output_len(req_id):
        _cur_out_len = SimulateVllm.req_to_output_len.get(req_id)
        if _cur_out_len is None:
            try:
                _cur_out_len = SimulateVllm.req_to_output_len.get(int(req_id))
            except ValueError:
                pass
        return _cur_out_len
 
    @staticmethod
    def update_token(model_input, sampling_output, eos_token_id=151645):
        # sampling output 是vllm.model_executor.layers.sampler.SamplerOutput 类
        for _com_seq_gp in sampling_output.outputs:
            # 每一条_com_seq_gp 对应的输出 _seq_out, _com_seq_gp 对应输入的index的输出。
            for _seq_out in _com_seq_gp.samples:
                # 当前output的对应的输入seq id。
                _p_s_id = _seq_out.parent_seq_id
                # 当前request id 可能会有多个seq，将所有seq的输出累加到一起作为这个request的输出。
                for _req_id, _seq_ids in model_input.request_ids_to_seq_ids.items():
                    # 当前输出是否不属于这个请求的，不处理。
                    if _p_s_id not in _seq_ids:
                        continue
                    _max_out_len = SimulateVllm.get_max_output_len(_req_id)
                    # 未找到采集时该请求的最大输出次数，不处理。
                    if _max_out_len is None:
                        continue
                    # 获取该请求当前
                    _cur_out_len = SimulateVllm.get_cur_output_len(_req_id)
                    if _cur_out_len is None:
                        continue
                    # 更新这个SequenceOutput的输出token。
                    _origin_token = _seq_out.output_token
                    if _cur_out_len < (
                            _max_out_len - 1) and _seq_out.output_token in SimulateVllm.req_to_stop_token_ids.get(
                            _req_id, [eos_token_id]):
                        _seq_out.output_token = np.random.randint(0, min(SimulateVllm.req_to_stop_token_ids))
                        _seq_out.logprobs[_seq_out.output_token] = _seq_out.logprobs.get(_origin_token)
                        _seq_out.logprobs.pop(_origin_token)
                    elif _cur_out_len >= (_max_out_len - 1):
                        _seq_out.output_token = SimulateVllm.req_to_stop_token_ids.get(_req_id, [eos_token_id])[0]
                        _seq_out.logprobs[_seq_out.output_token] = _seq_out.logprobs.get(_origin_token)
                        _seq_out.logprobs.pop(_origin_token)
                    # 处理完这个请求
                    break


# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import atexit
import json
import os
import re
import shlex
import shutil
import stat
import subprocess
import tempfile
import time
import xmlrpc.client
from copy import deepcopy
from math import exp, inf
from pathlib import Path
from typing import Any, Dict, List, Tuple, Optional
from xmlrpc.client import ServerProxy

import numpy as np
import pandas as pd
import psutil
from loguru import logger

from msserviceprofiler.modelevalstate.common import get_train_sub_path
from msserviceprofiler.modelevalstate.config.config import AnalyzeTool, BenchMarkConfig, MindieConfig, settings, \
    DeployPolicy, map_param_with_value, MODEL_EVAL_STATE_CONFIG_PATH, modelevalstate_config_path, \
    CUSTOM_OUTPUT, custom_output, BenchMarkPolicy
from msserviceprofiler.modelevalstate.config.config import default_support_field, PsoOptions, \
    PerformanceIndex, OptimizerConfigField
from msserviceprofiler.modelevalstate.inference.constant import IS_SLEEP_FLAG
from msserviceprofiler.modelevalstate.optimizer.analyze_profiler import analyze as analyze_profiler
from msserviceprofiler.modelevalstate.optimizer.global_best_custom import CustomGlobalBestPSO
from msserviceprofiler.modelevalstate.optimizer.store import DataStorage

_analyze_mapping = {
    AnalyzeTool.profiler.value: analyze_profiler
}


def kill_children(children):
    for child in children:
        if not child.is_running():
            continue
        try:
            child.send_signal(9)
            child.wait(10)
        except Exception as e:
            logger.error(f"Failed in kill the {child.pid} process. detail: {e}")
            continue
        if child.is_running():
            logger.error(f"Failed to kill the {child.pid} process.")


def kill_process(process_name):
    for proc in psutil.process_iter(["pid", "name"]):
        if not hasattr(proc, "info"):
            continue
        if process_name not in proc.info["name"]:
            continue
        children = psutil.Process(proc.pid).children(recursive=True)
        kill_children([proc])
        kill_children(children)


def remove_file(output_path: Path):
    if not output_path:
        return
    if not isinstance(output_path, Path):
        output_path = Path(output_path)
    if not output_path.exists():
        return
    if output_path.is_file():
        output_path.unlink()
        return
    for file in output_path.iterdir():
        if file.is_file():
            file.unlink()
        else:
            try:
                shutil.rmtree(file)
            except OSError:
                remove_file(file)


def backup(target, bak, class_name=""):
    if not target:
        return
    if not bak:
        return
    if not isinstance(target, Path):
        target = Path(target)
    if not isinstance(target, Path):
        bak = Path(bak)
    if not target.exists():
        return
    if not bak.exists():
        return
    new_file = bak.joinpath(class_name).joinpath(target.name)
    if target.is_file():
        new_file.parent.mkdir(parents=True, exist_ok=True)
        if not new_file.exists():
            shutil.copy(target, new_file)
    else:
        if new_file.exists():
            for child in new_file.iterdir():
                backup(child, new_file, class_name)
        else:
            shutil.copytree(target, new_file)


def close_file_fp(file_fp):
    if not file_fp:
        return
    try:
        # 检查file_fp是否是一个文件对象
        if hasattr(file_fp, 'close'):
            file_fp.close()
        else:
            # 如果file_fp是一个文件描述符，调用os.close()
            os.close(file_fp)
    except (AttributeError, OSError):
        return


@atexit.register
def clearing_residual_process():
    kill_process(MindieConfig().process_name)


class BenchMark:
    def __init__(self, benchmark_config: BenchMarkConfig, throughput_type: str = "common",
                 bak_path: Optional[Path] = None):
        self.benchmark_config = benchmark_config
        self.throughput_type = throughput_type
        self.bak_path = bak_path
        self.run_log = None
        self.run_log_offset = None
        self.run_log_fp = None
        self.process = None
        self.pattern = re.compile(r"\s*(\d+\.?\d*)\s*\%")

    def backup(self, del_log=True):
        backup(self.benchmark_config.output_path, self.bak_path, self.__class__.__name__)
        if not del_log:
            backup(self.run_log, self.bak_path, self.__class__.__name__)

    def get_performance_index(self):
        output_path = Path(self.benchmark_config.output_path)
        common_generate_speed = None
        first_token_time = None
        perf_generate_token_speed = None
        decode_time = None
        success_rate = None
        for file in output_path.iterdir():
            if "result_common" in file.name:
                try:
                    df = pd.read_csv(file)
                    if "OutputGenerateSpeed" in df.columns:
                        _generate_speed = df["OutputGenerateSpeed"][0]
                    else:
                        _generate_speed = df["GenerateSpeed"][0]
                    if isinstance(_generate_speed, str):
                        common_generate_speed = float(_generate_speed.split()[0])
                    elif isinstance(_generate_speed,
                                    (int, float, np.int64, np.int32, np.float64, np.float32, np.float16)):
                        common_generate_speed = _generate_speed
                    else:
                        raise TypeError(f"GenerateSpeed: {_generate_speed}, type: {type(_generate_speed)}")
                    req_returnd = df["Returned"][0]
                    if not req_returnd:
                        continue
                    _m_res = self.pattern.search(req_returnd)
                    if not _m_res:
                        continue
                    success_rate = float(_m_res.group(1)) / 100
                except (KeyError, AttributeError) as e:
                    logger.error(f"Failed in get GenerateSpeed. error: {e}")
                continue
            if "result_perf" in file.name:
                try:
                    df = pd.read_csv(file)
                    first_token_time = float(df["FirstTokenTime"][0].split()[0])
                    perf_generate_token_speed = float(df["GeneratedTokenSpeed"][0].split()[0])
                    decode_time = float(df["DecodeTime"][0].split()[0])
                except (AttributeError, KeyError) as e:
                    logger.error("Failed in get FirstTokenTime or GeneratedTokenSpeed. error: {}", e)
        if common_generate_speed is None and perf_generate_token_speed is None:
            raise ValueError("Not Found common_generate_speed or perf_generate_token_speed.")
        if first_token_time is None or decode_time is None:
            raise ValueError("Not Found first_token_time.")
        if self.throughput_type == "common":
            generate_speed = common_generate_speed
        else:
            generate_speed = perf_generate_token_speed
        time_to_first_token = first_token_time / 10 ** 3
        time_per_output_token = decode_time / 10 ** 3
        return PerformanceIndex(generate_speed=generate_speed,
                                time_to_first_token=time_to_first_token,
                                time_per_output_token=time_per_output_token,
                                success_rate=success_rate)

    def prepare(self):
        remove_file(Path(self.benchmark_config.output_path))
        remove_file(Path(self.benchmark_config.custom_collect_output_path))

    def check_success(self, print_log=False):
        if self.run_log:
            run_log_path = Path(self.run_log)
            if run_log_path.exists() and print_log:
                try:
                    with open(run_log_path, "r", encoding="utf-8") as f:
                        f.seek(self.run_log_offset)
                        output = f.read()
                        self.run_log_offset = f.tell()
                        logger.info(f"benchmark out: \n{output}")
                except (UnicodeError, OSError) as e:
                    logger.error(f"Failed read benchmark log. error {e}")
        try:
            if self.process.poll() is None:
                return False
            elif self.process.poll() == 0:
                return True
            else:
                raise subprocess.SubprocessError(
                    f"Failed in run benchmark. return code: {self.process.returncode}. ")
        except AttributeError as e:
            logger.error(f"Failed to check process status, error {e}")
            return False

    def run(self, run_params: Tuple[OptimizerConfigField]):
        # 启动测试
        logger.info("Start the benchmark test.")
        self.run_log_fp, self.run_log = tempfile.mkstemp(prefix="modelevalstate_benchmark")
        self.run_log_offset = 0
        if self.benchmark_config.work_path:
            cwd = self.benchmark_config.work_path
        else:
            cwd = os.getcwd()
        for k in run_params:
            if k.config_position == "env":
                try:
                    os.environ[k.name] = str(k.value)
                except KeyError as e:
                    logger.error(f"Failed to set environment variable. error {e}")
        if CUSTOM_OUTPUT not in os.environ:
            os.environ[CUSTOM_OUTPUT] = str(custom_output)
        run_cmd = shlex.split(self.benchmark_config.command)
        try:
            self.process = subprocess.Popen(run_cmd, env=os.environ, stdout=self.run_log_fp, stderr=subprocess.STDOUT,
                                            text=True, cwd=cwd)
        except OSError as e:
            logger.error(f"Failed to run benchmark. error {e}")
            raise e
        logger.info(f"command: {' '.join(run_cmd)}, log file: {self.run_log}")

    def stop(self, del_log=True):
        self.backup(del_log)
        close_file_fp(self.run_log_fp)
        try:
            if self.process and self.process.poll() is None:
                self.process.kill()
        except AttributeError as e:
            logger.error(f"Failed to kill process. error {e}")
        if del_log:
            remove_file(Path(self.run_log))


class ProfilerBenchmark(BenchMark):
    def __init__(self, benchmark_config: BenchMarkConfig, *args, analyze_tool: AnalyzeTool = AnalyzeTool.default,
                 **kwargs):
        super().__init__(benchmark_config, *args, **kwargs)
        self.analyze_tool = analyze_tool
        self.profiler_cmd = ["python", "-m", "ms_service_profiler.parse",
                             f"--input-path={self.benchmark_config.profile_input_path}",
                             f"--output-path={self.benchmark_config.profile_output_path}"]
        self.profiler_log = None
        self.profiler_log_fp = None
        self.profiler_log_offset = 0
        self.profiler_process = None

    def extra_performance_index(self, *args, **kwargs):
        logger.info("extra_performance_index")
        analyze_tool = _analyze_mapping.get(self.analyze_tool)
        if analyze_tool is None:
            raise ValueError(f"Analyze tool not found: {self.analyze_tool}")
        res = analyze_tool(*args, **kwargs)
        time_to_first_token = time_per_output_token = success_rate = None
        if isinstance(res, tuple):
            if len(res) == 1:
                generate_speed = res[0]
            elif len(res) == 2:
                generate_speed, time_to_first_token = res
            elif len(res) == 3:
                generate_speed, time_to_first_token, time_per_output_token = res
            elif len(res) == 4:
                generate_speed, time_to_first_token, time_per_output_token, success_rate = res
            else:
                raise ValueError(f"Not Support. res: {res}")
        else:
            generate_speed = res
        return PerformanceIndex(generate_speed=generate_speed, time_to_first_token=time_to_first_token,
                                time_per_output_token=time_per_output_token, success_rate=success_rate)


    def backup(self, del_log=True):
        super().backup(del_log)
        backup(self.benchmark_config.profile_input_path, self.bak_path, self.__class__.__name__)
        backup(self.benchmark_config.profile_output_path, self.bak_path, self.__class__.__name__)
        if not del_log and self.profiler_log:
            backup(self.profiler_log, self.bak_path, self.__class__.__name__)

    def prepare(self):
        super().prepare()
        remove_file(Path(self.benchmark_config.profile_input_path))
        remove_file(Path(self.benchmark_config.profile_output_path))

    def check_profiler(self, print_log=False):
        if print_log:
            try:
                with open(self.profiler_log, "r") as f:
                    f.seek(self.profiler_log_offset)
                    output = f.read()
                    self.profiler_log_offset = f.tell()
            except (UnicodeError, OSError) as e:
                logger.error(f"Failed read benchmark log. error {e}")
            if output:
                logger.info(f"benchmark out: \n{output}")
        if self.profiler_process.poll() is None:
            return False
        elif self.profiler_process.poll() == 0:
            return True
        else:
            raise subprocess.SubprocessError(
                f"Failed in run benchmark. return code: {self.process.returncode}. ")

    def start_profiler(self):
        self.profiler_log_fp, self.profiler_log = tempfile.mkstemp(prefix="modelevalstate_profiler")
        self.profiler_log_offset = 0
        if not os.path.exists(self.benchmark_config.work_path):
            raise FileNotFoundError(f"Work path not found: {self.benchmark_config.work_path}")
        logger.info(f"command: {' '.join(self.profiler_cmd)}, log file: {self.profiler_log}")
        self.profiler_process = subprocess.Popen(self.profiler_cmd, env=os.environ, stdout=self.profiler_log_fp,
                                                 stderr=subprocess.STDOUT,
                                                 text=True, cwd=self.benchmark_config.work_path)

    def get_performance_index(self):
        logger.info("get_performance_index")
        try:
            self.start_profiler()
            logger.info("wait profiler")
            while True:
                if self.check_profiler(print_log=True):
                    break
                time.sleep(1)
        except Exception as e:
            logger.error(f"Failed in start profiler. relation log: {self.profiler_log}")
            raise e
        collect_path = Path(self.benchmark_config.custom_collect_output_path)
        if self.analyze_tool == AnalyzeTool.profiler.value:
            res = self.extra_performance_index(self.benchmark_config.profile_output_path, collect_path)
            return res
        else:
            return super().get_performance_index()

    def stop(self, del_log=True):
        super().stop(del_log)
        close_file_fp(self.profiler_log_fp)
        if del_log:
            remove_file(Path(self.profiler_log))
        try:
            if self.profiler_process and self.profiler_process.poll() is None:
                self.profiler_process.kill()
        except AttributeError as e:
            logger.error(f"Failed to kill process. error {e}")


class VllmBenchMark(BenchMark):
    def __init__(self, benchmark_config: BenchMarkConfig, throughput_type: str = "common",
                 bak_path: Optional[Path] = None):
        super().__init__(benchmark_config, throughput_type, bak_path)
        self.output_path = benchmark_config.output_path
        if not self.output_path.exists():
            self.output_path.mkdir(parents=True)
 
    def get_performance_index(self):
        output_path = Path(self.benchmark_config.output_path)
        generate_speed = None
        time_per_output_token = None
        time_to_first_token = None
        success_rate = None
        for file in output_path.iterdir():
            if not file.name.endswith(".json"):
                continue
            with open(file, mode='r', encoding="utf-8") as f:
                data = json.load(f)
            generate_speed = data.get("output_throughput", 0)
            time_to_first_token = data.get("mean_ttft_ms", 0) / 10 ** 3
            time_per_output_token = data.get("mean_tpot_ms", 0) / 10 ** 3
            num_prompts = data.get("num_prompts", 1)
            completed = data.get("completed", 0)
            success_rate = completed / num_prompts
        return PerformanceIndex(generate_speed=generate_speed,
                                time_to_first_token=time_to_first_token,
                                time_per_output_token=time_per_output_token,
                                success_rate=success_rate)
 
    def run(self, run_params: Tuple[OptimizerConfigField]):
        # 启动测试
        logger.info("Start the benchmark test.")
        self.run_log_fp, self.run_log = tempfile.mkstemp(prefix="modelevalstate_benchmark")
        self.run_log_offset = 0
        if self.benchmark_config.work_path:
            cwd = self.benchmark_config.work_path
        else:
            cwd = os.getcwd()
        for k in run_params:
            if k.config_position == "env":
                try:
                    os.environ[k.name] = str(k.value)
                except KeyError as e:
                    logger.error(f"Failed to set environment variable. error {e}")
        if CUSTOM_OUTPUT not in os.environ:
            os.environ[CUSTOM_OUTPUT] = str(custom_output)
        os.environ["MODEL_EVAL_STATE_VLLM_CUSTOM_OUTPUT"] = str(self.output_path)
        run_cmd = shlex.split(self.benchmark_config.command)
        try:
            self.process = subprocess.Popen(run_cmd, env=os.environ, stdout=self.run_log_fp, stderr=subprocess.STDOUT,
                                            text=True, cwd=cwd)
        except OSError as e:
            logger.error(f"Failed to run benchmark. error {e}")
            raise e
        logger.info(f"command: {' '.join(run_cmd)}, log file: {self.run_log}")


class Simulator:
    def __init__(self, mindie_config: MindieConfig, bak_path: Optional[Path] = None):
        self.mindie_config = mindie_config
        logger.info(f"config path {self.mindie_config.config_path}", )
        if not self.mindie_config.config_path.exists():
            raise FileNotFoundError(self.mindie_config.config_path)
        with open(self.mindie_config.config_path, "r") as f:
            data = json.load(f)
        self.default_config = data
        logger.info(f"config bak path {self.mindie_config.config_bak_path}", )
        if not self.mindie_config.config_bak_path.exists():
            flags = os.O_WRONLY | os.O_CREAT
            modes = stat.S_IWUSR | stat.S_IRUSR
            with os.fdopen(os.open(self.mindie_config.config_bak_path, flags, modes), 'w') as fout:
                json.dump(self.default_config, fout, indent=4)
        self.mindie_log = None
        self.mindie_log_offset = 0
        self.bak_path = bak_path
        self.mindie_log_fp = None
        self.process = None

    @staticmethod
    def get_new_config(origin_config, params: Tuple[OptimizerConfigField], upper_key: str = "") -> Any:
        if upper_key:
            _keys = [upper_key]
        else:
            _keys = []
        if isinstance(origin_config, dict):
            _dict_config = {}
            for k, v in origin_config.items():
                _root_key = ".".join([*_keys, k])
                new_value = Simulator.get_new_config(v, params, _root_key)
                _dict_config[k] = new_value
            return _dict_config

        elif isinstance(origin_config, list):
            _list_config = []
            for i, v in enumerate(origin_config):
                _root_key = ".".join([*_keys, str(i)])
                new_value = Simulator.get_new_config(v, params, f"{upper_key}.{i}")
                _list_config.append(new_value)
            return _list_config
        else:
            for _p in params:
                if upper_key == _p.config_position:
                    logger.info(f"Update Config key: {upper_key}")
                    return _p.value
            return origin_config

    @staticmethod
    def set_config(origin_config, key: str, value: Any):
        next_level = None
        if "." in key:
            _f_index = key.index(".")
            _cur_key, next_level = key[:_f_index], key[_f_index + 1:]
        else:
            _cur_key = key
        if next_level:
            if isinstance(origin_config, dict):
                Simulator.set_config(origin_config[_cur_key], next_level, value)
            elif isinstance(origin_config, list):
                Simulator.set_config(origin_config[int(_cur_key)], next_level, value)
            else:
                raise ValueError(f"Not Support type {type(origin_config)}")
        else:
            origin_config[_cur_key] = value

    def backup(self, del_log=True):
        backup(self.mindie_config.config_path, self.bak_path, self.__class__.__name__)
        if not del_log and self.mindie_log:
            backup(self.mindie_log, self.bak_path, self.__class__.__name__)

    def update_config(self, params: Tuple[OptimizerConfigField]):
        # 将params值更新到新的config中
        new_config = deepcopy(self.default_config)
        for p in params:
            if not p.config_position.startswith("BackendConfig"):
                continue
            Simulator.set_config(new_config, p.config_position, p.value)

        # 将新的config写入到config文件中
        logger.debug(f"new config {new_config}")
        flags = os.O_WRONLY | os.O_CREAT
        modes = stat.S_IWUSR | stat.S_IRUSR
        if self.mindie_config.config_path.exists():
            self.mindie_config.config_path.unlink()
        with os.fdopen(os.open(self.mindie_config.config_path, flags, modes), "w") as fout:
            json.dump(new_config, fout, indent=4)

    def check_env(self):
        logger.info("check env")
        _residual_process = []
        _all_process_name = self.mindie_config.process_name.split(",")
        for proc in psutil.process_iter(["pid", "name"]):
            if not hasattr(proc, "info"):
                continue
            _proc_flag = []
            for p in _all_process_name:
                if p not in proc.info["name"]:
                    _proc_flag.append(True)
                else:
                    _proc_flag.append(False)
            if all(_proc_flag):
                continue
            _residual_process.append(proc)
        if _residual_process:
            logger.info("kill residual_process")
            for _p_name in _all_process_name:
                try:
                    kill_process(_p_name)
                except Exception as e:
                    logger.error(f"Failed to kill process. {e}")
        time.sleep(1)

    def check_success(self, print_log=False):
        with open(self.mindie_log, "r") as f:
            try:
                f.seek(self.mindie_log_offset)
                output = f.read()
                self.mindie_log_offset = f.tell()
            except Exception as e:
                logger.info(f"Failed in read mindie log. error: {e}")
        if output:
            if print_log:
                logger.info(f"simulate out: \n{output}")
            if "Daemon start success!" in output:
                return True
        if self.process.poll() is not None:
            raise subprocess.SubprocessError(
                f"Failed in run mindie. return code: {self.process.returncode}. "
                f"Please check the service log or console output.")
        return False

    def start_server(self, run_params: Tuple[OptimizerConfigField]):
        self.mindie_log_fp, self.mindie_log = tempfile.mkstemp(prefix="modelevalstate_mindie")
        self.mindie_log_offset = 0
        if self.mindie_config.work_path:
            cwd = self.mindie_config.work_path
        else:
            cwd = os.getcwd()
        for k in run_params:
            if k.config_position == "env":
                os.environ[k.name] = str(k.value)
        if MODEL_EVAL_STATE_CONFIG_PATH not in os.environ:
            os.environ[MODEL_EVAL_STATE_CONFIG_PATH] = str(modelevalstate_config_path)
        if CUSTOM_OUTPUT not in os.environ:
            os.environ[CUSTOM_OUTPUT] = str(custom_output)
        logger.debug(f"env {os.environ}")
        run_cmd = shlex.split(self.mindie_config.command)
        logger.info(f"run cmd: {run_cmd}, log path: {self.mindie_log}")
        self.process = subprocess.Popen(run_cmd, stdout=self.mindie_log_fp, stderr=subprocess.STDOUT, env=os.environ,
                                        text=True, cwd=cwd)

    def run(self, run_params: Tuple[OptimizerConfigField]):
        logger.info(f'start run in simulator. run params: {run_params}')
        # 根据params 修改配置文件
        self.update_config(run_params)
        # 启动mindie仿真
        try:
            self.check_env()
        except Exception as e:
            logger.error(f"Failed to check env. {e}")
        self.start_server(run_params)

    def stop(self, del_log=True):
        logger.info("Stop simulator process")
        if self.bak_path:
            self.backup()
        close_file_fp(self.mindie_log_fp)
        if del_log:
            remove_file(self.mindie_log)
        self.mindie_log_offset = 0
        if not self.process:
            return
        _process_state = self.process.poll()
        if _process_state is not None:
            logger.info(f"mindie already. exit_code: {_process_state}")
            return
        try:
            children = psutil.Process(self.process.pid).children(recursive=True)
            self.process.kill()
            try:
                self.process.wait(10)
            except subprocess.TimeoutExpired:
                self.process.send_signal(9)
            if self.process.poll() is not None:
                logger.info(f"The {self.process.pid} process has been shut down.")
            else:
                logger.error(f"The {self.process.pid} process shutdown failed.")
            kill_children(children)
            kill_process(self.mindie_config.process_name)
            remove_file(self.mindie_config.config_path)
            flags = os.O_WRONLY | os.O_CREAT
            modes = stat.S_IWUSR | stat.S_IRUSR
            with os.fdopen(os.open(self.mindie_config.config_path, flags, modes), "w") as fout:
                json.dump(self.default_config, fout)
        except Exception as e:
            logger.error(f"Failed to stop simulator process. {e}")


class VllmSimulator(Simulator):
    def __init__(self, mindie_config: MindieConfig, bak_path: Optional[Path] = None):
        try:
            super().__init__(mindie_config, bak_path)
        except Exception as e:
            logger.info('VllmSimulator init failed')
        self.mindie_config = mindie_config
        self.mindie_log = None
        self.mindie_log_offset = 0
        self.bak_path = bak_path
        self.mindie_log_fp = None
        self.process = None
 
    def run(self, run_params: Tuple[OptimizerConfigField]):
        logger.info(f'start run in simulator. run params: {run_params}')
        # 启动mindie仿真
        try:
            self.check_env()
        except Exception as e:
            logger.error(f"Failed to check env. {e}")
        self.start_server(run_params)
 
    def check_success(self, print_log=False):
        with open(self.mindie_log, "r") as f:
            try:
                f.seek(self.mindie_log_offset)
                output = f.read()
                self.mindie_log_offset = f.tell()
            except Exception as e:
                logger.info(f"Failed in read vllm log. error: {e}")
        if output:
            if print_log:
                logger.info(f"simulate out: \n{output}")
            if "Application startup complete." in output:
                return True
        if self.process.poll() is not None:
            raise subprocess.SubprocessError(
                f"Failed in run vllm. return code: {self.process.returncode}. "
                f"Please check the service log or console output.")
        return False
 
    def stop(self, del_log=True):
        logger.info("Stop simulator process")
        if self.bak_path:
            self.backup()
        close_file_fp(self.mindie_log_fp)
        if del_log:
            remove_file(self.mindie_log)
        self.mindie_log_offset = 0
        if not self.process:
            return
        _process_state = self.process.poll()
        if _process_state is not None:
            logger.info(f"vllm already. exit_code: {_process_state}")
            return
        try:
            children = psutil.Process(self.process.pid).children(recursive=True)
            self.process.kill()
            try:
                self.process.wait(10)
            except subprocess.TimeoutExpired:
                self.process.send_signal(9)
            if self.process.poll() is not None:
                logger.info(f"The {self.process.pid} process has been shut down.")
            else:
                logger.error(f"The {self.process.pid} process shutdown failed.")
            kill_children(children)
            kill_process(self.mindie_config.process_name)
        except Exception as e:
            logger.error(f"Failed to stop simulator process. {e}")


class Scheduler:
    def __init__(self, simulator: Simulator, benchmark: BenchMark, data_storage: DataStorage,
                 bak_path: Optional[Path] = None, retry_number: int = 3, wait_start_time=1800):
        self.simulator = simulator
        self.benchmark = benchmark
        self.data_storage = data_storage
        self.bak_path = bak_path
        self.retry_number = retry_number
        self.wait_time = wait_start_time
        self.current_back_path = None
        self.simulate_run_info = None

    def back_up(self):
        if self.bak_path:
            self.current_back_path = get_train_sub_path(self.bak_path)
            self.simulator.bak_path = self.current_back_path
            self.benchmark.bak_path = self.current_back_path

    def wait_simulate(self):
        logger.info("wait run mindie")
        for _ in range(self.wait_time):
            time.sleep(1)
            if self.simulator.check_success():
                logger.info(f"Successfully started the {self.simulator.process.pid} process.")
                return
        raise TimeoutError(self.wait_time)


    def run_simulate(self, params: np.ndarray, params_field: Tuple[OptimizerConfigField]):
        self.benchmark.prepare()
        self.simulator.run(tuple(self.simulate_run_info))
        self.wait_simulate()

    def monitoring_status(self):
        logger.info("monitor status")
        while True:
            if self.simulator.process.poll() is not None:
                self.simulator.stop(del_log=False)
                self.benchmark.stop(del_log=False)
                raise subprocess.SubprocessError(f"Failed in run simulator. "
                                                 f"return code: {self.simulator.process.returncode}.")
            if self.benchmark.check_success():
                return
            time.sleep(1)

    def run_target_server(self, params: np.ndarray, params_field: Tuple[OptimizerConfigField]):
        """
        1. 启动mindie仿真
        2. 启动benchmark 测试
        3. 检查mindie状态，检查benchmark状态
        """
        for _ in range(self.retry_number):
            try:
                self.run_simulate(params, params_field)
            except Exception as e:
                logger.error(f"Failed in Simulator Running. error: {e}， simulator log {self.simulator.mindie_log}")
                logger.exception("What?!")
                self.stop_target_server(del_log=False)
                continue
            time.sleep(1)
            try:
                self.benchmark.run(tuple(self.simulate_run_info))
            except Exception as e:
                logger.error(f"Failed in Benchmark Running. error: {e}, benchmark log {self.benchmark.run_log}")
                logger.exception("What?!")
                self.stop_target_server(del_log=False)
                continue
            time.sleep(1)
            try:
                self.monitoring_status()
            except Exception as e:
                self.stop_target_server(del_log=False)
                logger.error(f"Failed in monitoring status. error: {e}, simulator log {self.simulator.mindie_log}, "
                             f"benchmark log {self.benchmark.run_log}")
                logger.exception("What?!")
                continue
            return
        raise ValueError(
            f"Failed in run_target_server")

    def stop_target_server(self, del_log=True):
        self.simulator.stop(del_log)
        self.benchmark.stop(del_log)

    def run(self, params: np.ndarray, params_field: Tuple[OptimizerConfigField]) -> PerformanceIndex:
        """
        1. 启动mindie仿真
        2. 启动benchmark 测试
        3. 获取benchmark测试结果
        4. 关闭mindie仿真
        5. 返回benchmark测试结果
        params: 是一维数组，其值对应mindie 的相关配置。
        """
        logger.info("Start run in scheduler.")
        self.back_up()
        self.simulate_run_info = map_param_with_value(params, params_field)
        error_info = None
        del_log = True
        performance_index = PerformanceIndex()
        try:
            self.run_target_server(params, params_field)
            time.sleep(1)
            performance_index = self.benchmark.get_performance_index()
        except Exception as e:
            logger.error(f"Failed running. bak path: {self.simulator.bak_path}")
            error_info = e
            del_log = False
        self.data_storage.save(performance_index, tuple(self.simulate_run_info), self.benchmark.benchmark_config,
                               error=error_info, bakcup=self.current_back_path)
        self.stop_target_server(del_log)
        if error_info:
            raise error_info
        return performance_index


class ScheduleWithMultiMachine(Scheduler):
    def __init__(self, rpc_clients: List[ServerProxy], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rpc_clients = rpc_clients

    def back_up(self):
        if self.bak_path:
            _cur_bak_path = get_train_sub_path(self.bak_path)
            self.simulator.bak_path = _cur_bak_path
            self.benchmark.bak_path = _cur_bak_path
            for rpc in self.rpc_clients:
                if rpc.simulator:
                    rpc.simulator.bak_path = _cur_bak_path

    def monitoring_status(self):
        logger.info("Start monitoring")
        while True:
            all_poll = [self.simulator.process.poll()]
            for rpc in self.rpc_clients:
                all_poll.append(rpc.process_poll())
            if any([_i is not None for _i in all_poll]):
                self.stop_target_server(del_log=False)
                raise subprocess.SubprocessError(
                    f"Failed in run simulator. all status: {all_poll}, machine info: master, {self.rpc_clients}.")
            if self.benchmark.check_success():
                return
            time.sleep(1)

    def run_simulate(self, params: np.ndarray, params_field: Tuple[OptimizerConfigField]):
        self.benchmark.prepare()
        _simulate_run_info = map_param_with_value(params, params_field)
        [rpc.run_simulator(params.tolist()) for rpc in self.rpc_clients]
        self.simulator.run(tuple(self.simulate_run_info))
        self.wait_simulate()
        [rpc.check_success() for rpc in self.rpc_clients]

    def stop_target_server(self, del_log=True):
        super(ScheduleWithMultiMachine, self).stop_target_server(del_log)
        [rpc.stop_simulator(del_log) for rpc in self.rpc_clients]


class PSOOptimizer:
    def __init__(self, scheduler: Scheduler, n_particles: int = 10, iters=100, pso_options: PsoOptions = None,
                 target_field: Optional[Tuple] = None, prefill_lam: float = 0.5, decode_lam: float = 0.5,
                 success_rate_lam: float = 0.5, prefill_constrain: float = 0.05, decode_constrain: float = 0.05,
                 success_rate_constrain: float = 1, load_history_data: Optional[List] = None,
                 load_breakpoint: bool = False, pso_init_kwargs: Optional[Dict] = None):
        self.scheduler = scheduler
        self.n_particles = n_particles
        self.iters = iters
        self.target_field = target_field if target_field else default_support_field
        if not pso_options:
            self.pso_options = PsoOptions()
        else:
            self.pso_options = pso_options
        self.prefill_lam = prefill_lam  # 优化算法中惩罚系数
        self.decode_lam = decode_lam
        self.success_rate_lam = success_rate_lam
        self.prefill_constrain = prefill_constrain
        self.decode_constrain = decode_constrain
        self.success_rate_constrain = success_rate_constrain
        self.load_history_data = load_history_data
        self.load_breakpoint = load_breakpoint
        self.pso_init_kwargs = pso_init_kwargs
        self.init_pos = None
        self.history_cost, self.history_pos = None, None
        if self.load_history_data and self.load_breakpoint:
            self.history_pos, self.history_cost = self.computer_fitness()

    def computer_fitness(self) -> Tuple:
        all_position = []
        all_cost = []
        for case_data in self.load_history_data:
            _params = {}
            for k in PerformanceIndex.model_fields.keys():
                if k in case_data:
                    _params[k] = case_data[k]
            performance_index = PerformanceIndex(**_params)
            try:
                _fitness = self.minimum_algorithm(performance_index)
                logger.info(f"fitness {_fitness}")
                _pos = [case_data.get(_field.name) for _field in self.target_field]
                if not all(_pos):
                    continue
                all_cost.append(_fitness)
                all_position.append(_pos)
            except KeyError:
                continue
        if len(all_position) != len(all_cost):
            raise ValueError("Failed in computer_fitness.")
        return all_position, all_cost

    def minimum_algorithm(self, performance_index: PerformanceIndex) -> float:
        try:
            fitness = 1 / performance_index.generate_speed
        except OverflowError:
            return inf
        if performance_index.time_to_first_token is not None:
            _var = max(0.0, (
                    performance_index.time_to_first_token - self.prefill_constrain) / self.prefill_constrain)
            try:
                fitness += self.prefill_lam * (exp(_var) - 1)
            except OverflowError:
                return inf
        if performance_index.time_per_output_token is not None:
            _decode_var = max(0.0, (
                    performance_index.time_per_output_token - self.decode_constrain) / self.decode_constrain)
            try:
                fitness += self.decode_lam * (exp(_decode_var) - 1)
            except OverflowError:
                return inf
        if performance_index.success_rate:
            _success_var = max(0.0, (
                    performance_index.success_rate - self.success_rate_constrain) / self.success_rate_constrain)
            try:
                fitness += self.success_rate_lam * (exp(_success_var) - 1)
            except OverflowError:
                return inf
        return fitness

    def op_func(self, x) -> np.ndarray:
        n_particles = x.shape[0]
        logger.info(f"Acquired n_particles: {n_particles}, value: {x}")
        generate_speed = []
        for i in range(n_particles):
            # 调用schedule， 获取采集的数据
            try:
                _res = self.scheduler.run(x[i], self.target_field)
                # 根据采集的数据，计算最优化值
                _fitness = self.minimum_algorithm(_res)
            except Exception as e:
                logger.error(f"Failed. error: {e}, please check.")
                logger.exception("What?!")
                _fitness = inf
            logger.info(f"fitness {_fitness}")
            generate_speed.append(_fitness)
        return np.array(generate_speed)

    def constructing_bounds(self) -> Tuple[Tuple, Tuple]:
        """
        返回示例：((0, 10), (0, 10))
        """
        _min = []
        _max = []
        for _field in self.target_field:
            _min.append(_field.min)
            _max.append(_field.max)
        return (tuple(_min), tuple(_max))

    def run(self):
        optimizer = CustomGlobalBestPSO(n_particles=self.n_particles, dimensions=len(self.target_field),
                                        options=self.pso_options.model_dump(), bounds=self.constructing_bounds(),
                                        init_pos=self.init_pos, breakpoint_pos=self.history_pos,
                                        breakpoint_cost=self.history_cost, **self.pso_init_kwargs)
        cost, joint_vars = optimizer.optimize(self.op_func, iters=self.iters)
        best_position = {_field.name: _field.value for _field in map_param_with_value(joint_vars, self.target_field)}
        logger.info(f"best cost {cost}, best joint_vars: {best_position}")


def arg_parse(subparsers):
    parser = subparsers.add_parser(
        "optimizer", formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="optimize for performance"
    )

    parser.add_argument("-lb", "--load_breakpoint", action="store_true",
                        help="Continue from where the last optimization was aborted.")
    parser.add_argument("-d", "--deploy_policy", default=DeployPolicy.single.value,
                        choices=[k.value for k in list(DeployPolicy)],
                        help="Indicates whether the multi-node running policy is used.")
    parser.add_argument("--backup", default=False, action="store_true",
                        help="Whether to back up data.")
    parser.add_argument("-b", "--benchmark_policy", default=BenchMarkPolicy.benchmark.value,
                        choices=[k.value for k in list(BenchMarkPolicy)],
                        help="Whether to use custom performance indicators.")
    parser.set_defaults(func=main)


def main(args: argparse.Namespace):
    if args.benchmark_policy == BenchMarkPolicy.vllm_benchmark.value:
        simulator = VllmSimulator(settings.simulator)
    else:
        simulator = Simulator(settings.simulator)
    bak_path = None
    if args.backup:
        bak_path = settings.output.joinpath("bak")
        if not bak_path.exists():
            bak_path.mkdir(parents=True)
    rpc_clients = []
    if args.deploy_policy == DeployPolicy.multiple.value:
        for server_address in settings.server:
            rpc = xmlrpc.client.ServerProxy(server_address, allow_none=True)
            logger.info(f"{server_address} support method {rpc.system.listMethods()}")
            rpc_clients.append(rpc)
    # 单机benchmark
    time_sleep = os.getenv(IS_SLEEP_FLAG, "False").lower().strip() == "true"
    if time_sleep and args.benchmark_policy == BenchMarkPolicy.benchmark.value:
        benchmark = BenchMark(settings.benchmark, bak_path=bak_path)
    elif time_sleep and args.benchmark_policy == BenchMarkPolicy.vllm_benchmark.value:
        benchmark = VllmBenchMark(settings.benchmark, bak_path=bak_path)
    else:
        # 默认 自定义单机
        benchmark = ProfilerBenchmark(settings.benchmark, bak_path=bak_path, analyze_tool=AnalyzeTool.profiler)
    # 存储结果，只在主节点存储结果
    data_storage = DataStorage(settings.data_storage)
    # 初始化调度模块，支持单机和多机。
    if args.deploy_policy == DeployPolicy.multiple.value:
        scheduler = ScheduleWithMultiMachine(rpc_clients, simulator, benchmark, data_storage, bak_path=bak_path)
    else:
        scheduler = Scheduler(simulator, benchmark, data_storage, bak_path=bak_path)
    _load_history_data = None
    _load_history = None
    if args.load_breakpoint:
        _load_history = True
    if _load_history:
        _load_history_data = data_storage.load_history_position(settings.data_storage.store_dir)
    pso = PSOOptimizer(scheduler, n_particles=settings.n_particles, iters=settings.iters,
                       prefill_lam=settings.prefill_lam, target_field=settings.target_field,
                       decode_lam=settings.decode_lam, success_rate_lam=settings.success_rate_lam,
                       decode_constrain=settings.decode_constrain, prefill_constrain=settings.prefill_constrain,
                       success_rate_constrain=settings.success_rate_constrain, load_history_data=_load_history_data,
                       load_breakpoint=args.load_breakpoint,
                       pso_init_kwargs={"ftol": settings.ftol, "ftol_iter": settings.ftol_iter})
    pso.run()



# !/usr/bin/python3.8
# -*- coding: utf-8 -*-
# Copyright (c) Huawei Technologies Co., Ltd. 2023-2024. All rights reserved.
import os
import time
from copy import deepcopy
from enum import Enum
from pathlib import Path
from typing import Any, List, Tuple, Type, Optional, Union

import numpy as np
from loguru import logger
from pydantic import BaseModel, field_validator, Field
from pydantic_settings import BaseSettings, SettingsConfigDict, PydanticBaseSettingsSource, JsonConfigSettingsSource

import msserviceprofiler.modelevalstate

RUN_TIME = time.strftime("%Y%m%d%H%M%S", time.localtime())
INSTALL_PATH = Path(msserviceprofiler.modelevalstate.__path__[0])
RUN_PATH = Path(os.getcwd())
MODEL_EVAL_STATE_CONFIG_PATH = "MODEL_EVAL_STATE_CONFIG_PATH"
modelevalstate_config_path = os.getenv(MODEL_EVAL_STATE_CONFIG_PATH) or os.getenv(MODEL_EVAL_STATE_CONFIG_PATH.lower())
if not modelevalstate_config_path:
    modelevalstate_config_path = RUN_PATH.joinpath("config.json")
modelevalstate_config_path = Path(modelevalstate_config_path).absolute().resolve()

CUSTOM_OUTPUT = "MODEL_EVAL_STATE_CUSTOM_OUTPUT"
custom_output = os.getenv(CUSTOM_OUTPUT) or os.getenv(CUSTOM_OUTPUT.lower())
if custom_output:
    custom_output = Path(custom_output).resolve()
else:
    custom_output = RUN_PATH


class AnalyzeTool(Enum):
    default = "default"
    profiler = "profiler"
    vllm_benchmark = "vllm"


class BenchMarkPolicy(Enum):
    benchmark = "benchmark"
    profiler_benchmark = "profiler_benchmark"
    vllm_benchmark = "vllm_benchmark"


class DeployPolicy(Enum):
    single = "single"
    multiple = "multiple"


class OptimizerConfigField(BaseModel):
    name: str = "max_batch_size"
    config_position: str = "BackendConfig.ScheduleConfig.maxBatchSize"
    min: int = 0
    max: int = 100
    dtype: str = "float"
    value: Union[int, float, bool] = 0.0
    dtype_param: Any = None


def map_param_with_value(params: np.ndarray, params_field: Tuple[OptimizerConfigField]):
    _simulate_run_info = []
    for i, v in enumerate(params_field):
        _field = deepcopy(v)
        if v.dtype == "int":
            _field.value = int(params[i])
        elif v.dtype == "bool":
            if params[i] > 0.5:
                _field.value = True
            else:
                _field.value = False
        elif v.dtype == "enum":
            segment = np.linspace(v.min, v.max, len(v.dtype_param) + 1)
            if params[i] <= v.min:
                _field.value = v.dtype_param[0]
            elif params[i] >= v.max:
                _field.value = v.dtype_param[-1]
            else:
                _enum_index = np.searchsorted(segment, params[i]) - 1
                _field.value = v.dtype_param[_enum_index]
        else:
            _field.value = float(params[i])
        _simulate_run_info.append(_field)
    for i, v in enumerate(params_field):
        if v.dtype == "ratio":
            _field = _simulate_run_info[i]
            _t_op = [_op for _op in _simulate_run_info if _op.name == v.dtype_param][0]
            _field.value = int(_field.value * _t_op.value)
        if v.dtype == "factories":
            _field = _simulate_run_info[i]
            _t_op = [_op for _op in _simulate_run_info if _op.name == v.dtype_param["target_name"]][0]
            _field.value = eval(v.dtype_param["dtype"])(v.dtype_param["product"] / _t_op.value)

    return _simulate_run_info


class PerformanceIndex(BaseModel):
    generate_speed: Optional[float] = None
    time_to_first_token: Optional[float] = None
    time_per_output_token: Optional[float] = None
    success_rate: Optional[float] = None


class BenchMarkConfig(BaseModel):
    name: str = "benchmark"
    work_path: Path = Field(default_factory=lambda: Path(os.getcwd()).resolve())
    command: str = "/usr/bin/bash ./run_benchmark.sh"
    output_path: Path = custom_output.joinpath("instance")
    custom_collect_output_path: Path = Field(
        default_factory=lambda: custom_output.joinpath("result/custom_collect_output_path").resolve(),
        validate_default=True
    )
    profile_input_path: Path = Field(
        default_factory=lambda: custom_output.joinpath("result/profile_input_path").resolve(),
        validate_default=True
    )
    profile_output_path: Path = Field(
        default_factory=lambda: custom_output.joinpath("result/profile_output_path").resolve(),
        validate_default=True
    )

    @field_validator("output_path", "custom_collect_output_path", "profile_input_path",
                     "profile_output_path")
    @classmethod
    def create_path(cls, path: Path) -> Path:
        path.mkdir(parents=True, exist_ok=True)
        return path

    @field_validator("work_path")
    @classmethod
    def check_dir(cls, path: Path) -> Path:
        if not path.exists():
            logger.error(f"FileNotFound: {path}")
        return path


class DataStorageConfig(BaseModel):
    store_dir: Path = Field(
        default_factory=lambda: custom_output.joinpath("result/store").resolve(),
        validate_default=True
    )

    @field_validator("store_dir")
    @classmethod
    def create_path(cls, path: Path) -> Path:
        path.mkdir(parents=True, exist_ok=True)
        return path
    

class LatencyModel(BaseModel):
    base_path: Path = Field(default_factory=lambda: custom_output.joinpath("result/latency_model").resolve())
    model_path: Optional[Path] = Field(
        default_factory=lambda data: data["base_path"].joinpath("bak/base/xgb_model.ubj").resolve())
    ohe_path: Optional[Path] = Field(default_factory=lambda data: data["base_path"].joinpath("ohe").resolve())
    static_file_dir: Optional[Path] = Field(
        default_factory=lambda data: data["base_path"].joinpath("model_static_file").resolve(), validate_default=True)
    req_and_decode_file: Optional[Path] = Field(
        default_factory=lambda data: data["base_path"].joinpath("req_id_and_decode_num.json").resolve())
    cache_data: Optional[Path] = Field(
        default_factory=lambda data: data["base_path"].joinpath("cache").resolve())

    @field_validator("static_file_dir")
    @classmethod
    def validate_cache_data(cls, path: Optional[Path] = None):
        if path:
            if not Path(path).exists():
                path.mkdir(parents=True)
        return path
    
    @field_validator("static_file_dir")
    @classmethod
    def validate_static_file_dir(cls, path: Optional[None]):
        if path is None:
            path = cls.base_path.joinpath("model_static_file")
            path.mkdir(parents=True)
        return path


class MindieConfig(BaseModel):
    # 运行mindie时，要修改的mindie config
    process_name: str = "mindieservice_daemon"
    config_path: Path = Path("/usr/local/Ascend/mindie/latest/mindie-service/conf/config.json")
    config_bak_path: Path = Path("/usr/local/Ascend/mindie/latest/mindie-service/conf/config_bak.json")
    work_path: Path = Field(default_factory=lambda: Path(os.getcwd()).resolve())
    command: str = "/usr/bin/bash ./run_mindie.sh"
    log_path: Path = Path("/usr/local/Ascend/mindie/latest/mindie-service/logs")


    @field_validator("config_path")
    @classmethod
    def check_dir(cls, path: Path) -> Path:
        if not path.exists():
            logger.error(f"FileNotFound: {path}")
        return path

    @field_validator("log_path")
    @classmethod
    def create_path(cls, path: Path) -> Path:
        path.mkdir(parents=True, exist_ok=True)
        return path


class PsoOptions(BaseModel):
    c1: float = 2.0
    c2: float = 2.0
    w: float = 1.8


class PsoStrategy(BaseModel):
    w: str = "exp_decay"
    c1: str = "exp_decay"
    c2: str = "exp_decay"


default_support_field = [
    # max batch size 最小值要大于max_prefill_batch_size的最大值。
    OptimizerConfigField(name="max_batch_size", config_position="BackendConfig.ScheduleConfig.maxBatchSize", min=25,
                         max=300, dtype="int"),
    OptimizerConfigField(name="max_prefill_batch_size",
                         config_position="BackendConfig.ScheduleConfig.maxPrefillBatchSize", min=1, max=25,
                         dtype="int"),
    OptimizerConfigField(name="prefill_time_ms_per_req",
                         config_position="BackendConfig.ScheduleConfig.prefillTimeMsPerReq", max=1000, dtype="int"),
    OptimizerConfigField(name="decode_time_ms_per_req",
                         config_position="BackendConfig.ScheduleConfig.decodeTimeMsPerReq", max=1000, dtype="int"),
    OptimizerConfigField(name="support_select_batch",
                         config_position="BackendConfig.ScheduleConfig.supportSelectBatch", max=1,
                         dtype="bool"),
    OptimizerConfigField(name="max_prefill_token",
                         config_position="BackendConfig.ScheduleConfig.maxPrefillTokens", min=4096, max=409600,
                         dtype="int"),
    OptimizerConfigField(name="max_queue_deloy_mircroseconds",
                         config_position="BackendConfig.ScheduleConfig.maxQueueDelayMicroseconds", min=500, max=1000000,
                         dtype="int"),
    OptimizerConfigField(name="prefill_policy_type",
                         config_position="BackendConfig.ScheduleConfig.prefillPolicyType", min=0, max=1,
                         dtype="enum", dtype_param=[0, 1, 3]),
    OptimizerConfigField(name="decode_policy_type",
                         config_position="BackendConfig.ScheduleConfig.decodePolicyType", min=0, max=1,
                         dtype="enum", dtype_param=[0, 1, 3]),
    OptimizerConfigField(name="max_preempt_count",
                         config_position="BackendConfig.ScheduleConfig.maxPreemptCount", min=0, max=1,
                         dtype="ratio", dtype_param="max_batch_size")
]


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        json_file=[INSTALL_PATH.joinpath("model_eval_state.json"), Path("~/model_eval_state.json").expanduser(),
                   RUN_PATH.joinpath("model_eval_state.json"),
                   INSTALL_PATH.joinpath("config.json"), Path("~/config.json").expanduser(),
                   RUN_PATH.joinpath("config.json"), modelevalstate_config_path],
        env_prefix="model_eval_state_")
    output: Path = Field(default_factory=lambda: custom_output.joinpath("result").resolve(), validate_default=True)
    latency_model: LatencyModel = LatencyModel()
    simulator: MindieConfig = MindieConfig()
    benchmark: BenchMarkConfig = BenchMarkConfig()
    data_storage: DataStorageConfig = DataStorageConfig()
    pso_options: PsoOptions = PsoOptions()
    pso_strategy: PsoStrategy = PsoStrategy()
    n_particles: int = 5
    iters: int = 10
    ftol: float = -np.inf
    ftol_iter: int = 1
    prefill_lam: float = 0.5  # 惩罚系数
    decode_lam: float = 0.5
    success_rate_lam: float = 0.5
    prefill_constrain: float = 0.05
    decode_constrain: float = 0.05
    success_rate_constrain: float = 1.0
    server: List[str] = ["http://127.0.0.1:7425/"]
    float_range_in_best_particle: float = 0.1  # 如果用历史值作为作为初始值，那么允许在初始值的浮动程度。
    target_field: List[OptimizerConfigField] = default_support_field

    @classmethod
    def settings_customise_sources(
            cls,
            settings_cls: Type[BaseSettings],
            init_settings: PydanticBaseSettingsSource,
            env_settings: PydanticBaseSettingsSource,
            dotenv_settings: PydanticBaseSettingsSource,
            file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        return (init_settings, env_settings, JsonConfigSettingsSource(settings_cls), file_secret_settings)

    @field_validator("output")
    @classmethod
    def create_path(cls, path: Path) -> Path:
        path.mkdir(parents=True, exist_ok=True)
        return path


settings = Settings()



# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ms_service_profiler.exporters.base import ExporterBase
from ms_service_profiler.exporters.utils import save_dataframe_to_csv
from ms_service_profiler.utils.log import logger

from ..common.split_utils import get_batch_all_time, process_exporter, get_filter_df
from ..common.split_utils import get_statistics_data, preprocess_framework_df


class ExporterDecode(ExporterBase):
    name = "decode_data"
 
    @classmethod
    def initialize(cls, args):
        cls.args = args
 
    @classmethod
    def export(cls, data) -> None:
        output = cls.args.output_path
        log_level = cls.args.log_level
        batch_size = cls.args.decode_batch_size
        batch_num = cls.args.decode_number
        rid = cls.args.decode_rid
        df = data.get('tx_data_df')
        if df is None:
            logger.error("The data is empty, please check")
            return
        framework_df = preprocess_framework_df(df)
        if framework_df is None:
            return
        filter_df = get_filter_df(framework_df, 'Decode')
        add_all_time_df = get_batch_all_time(filter_df, 'Decode')
        framework_df = process_exporter(add_all_time_df, batch_size, batch_num, rid, 'Decode')
        if log_level == 'debug':
            save_dataframe_to_csv(add_all_time_df, output, "decode1.csv")
            save_dataframe_to_csv(framework_df, output, f"decode_{batch_num}.csv")
        framework_df = get_statistics_data(framework_df, 'batchFrameworkProcessing', 'Decode')
        if not framework_df.empty:
            save_dataframe_to_csv(framework_df, output, "decode.csv")
        logger.info("Export decode data successfully.")



# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import sys
import time
from typing import Tuple
from pathlib import Path

import xmlrpc.client
from xmlrpc.server import SimpleXMLRPCServer
from xmlrpc.server import SimpleXMLRPCRequestHandler

import numpy as np
from loguru import logger

from msserviceprofiler.modelevalstate.optimizer.optimizer import Simulator, remove_file
from msserviceprofiler.modelevalstate.config.config import settings, map_param_with_value


def get_file(target_path, parent_name: str = "", save_current_path: bool = False):
    # 获取该目录下所有文件,或者这个文件
    _work_dir = Path(target_path)
    if not _work_dir.exists():
        raise FileNotFoundError
    res = []

    if _work_dir.is_file():
        _file_name = parent_name + "/" + _work_dir.name if parent_name else _work_dir.name
        with open(_work_dir, "rb") as handle:
            res.append((_file_name, xmlrpc.client.Binary(handle.read())))
    else:
        if save_current_path:
            parent_name = parent_name + "/" + _work_dir.name if parent_name else _work_dir.name
        for child in _work_dir.iterdir():
            res.extend(get_file(child, parent_name, True))
    return res


# 限制为特定的路径。
class RequestHandler(SimpleXMLRPCRequestHandler):
    rpc_paths = ('/RPC2',)


class RemoteScheduler:
    def __init__(self):
        self.simulator = None

    def run_simulator(self, params: np.ndarray):
        # 更新服务器上的config文件
        self.simulator = Simulator(settings.mindie)
        _simulate_run_info = map_param_with_value(params, settings.target_field)
        logger.info(f"simulate run info {_simulate_run_info}")
        self.simulator.run(tuple(_simulate_run_info))

    def check_success(self):
        if not self.simulator:
            return None
        for _ in range(10):
            if self.simulator.check_success():
                return True
            time.sleep(10)
        raise Exception(f"Simulator run failed. please check log: {self.simulator.mindie_log}")

    def stop_simulator(self, del_log=True):
        if self.simulator:
            self.simulator.stop(del_log)

    def process_poll(self):
        if self.simulator:
            return self.simulator.process.poll()
        return None





# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import argparse
from pathlib import Path
import pandas as pd

from ms_service_profiler.exporters.utils import check_input_path_valid, check_output_path_valid


def add_exporters(args):
    from msserviceprofiler.ms_service_profiler_ext.exporters.exporter_prefill import ExporterPrefill
    from msserviceprofiler.ms_service_profiler_ext.exporters.exporter_decode import ExporterDecode

    if not hasattr(args, 'format'):
        args.format = 'csv'
    exporter_cls = []
    exporters = []
    if args.prefill_batch_size > 0 or args.prefill_rid != '-1':
        exporter_cls.append(ExporterPrefill)
    if args.decode_batch_size > 0 or args.decode_rid != '-1':
        exporter_cls.append(ExporterDecode)
    for cls in exporter_cls:
        exporter = cls()
        exporter.initialize(args)
        exporters.append(exporter)
    return exporters


def arg_parse(subparsers):
    parser = subparsers.add_parser(
        "split", formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="MS Server Profiler Split"
    )
    parser.add_argument(
        '--input-path', required=True, type=check_input_path_valid, help='Path to the folder containing profile data.'
    )
    parser.add_argument(
        '--output-path',
        type=check_output_path_valid,
        default=os.path.join(os.getcwd(), 'output'),
        help='Output file path to save results.')
    parser.add_argument(
        '--log-level',
        type=str,
        default='info',
        choices=['debug', 'info', 'warning', 'error', 'fatal', 'critical'],
        help='Log level to print')
    parser.add_argument('--prefill-batch-size', type=int, default=0, help='Batch size for Prefill data.')
    parser.add_argument('--decode-batch-size', type=int, default=0, help='Batch size for Decode data.')
    parser.add_argument(
        '--prefill-number', type=int, default=1, help='The number of Prefill batch to calc statistical data'
    )
    parser.add_argument(
        '--decode-number', type=int, default=1, help='The number of Decode batch to calc statistical data'
    )
    parser.add_argument('--prefill-rid', type=str, default='-1', help='The rid for Prefill batch to split')
    parser.add_argument('--decode-rid', type=str, default='-1', help='The rid for Decode batch to split')
    parser.set_defaults(func=main)


def main(args):
    from ms_service_profiler.parse import parse, preprocess_prof_folders
    from ms_service_profiler.utils.log import set_log_level
    from ms_service_profiler.plugins import custom_plugins

    # 初始化日志等级
    set_log_level(args.log_level)

    # msprof预处理
    preprocess_prof_folders(args.input_path)

    # 初始化Exporter
    exporters = add_exporters(args)

    # 创建output目录
    Path(args.output_path).mkdir(parents=True, exist_ok=True)

    # 解析数据并导出
    parse(args.input_path, custom_plugins, exporters, args=args)



# -*- coding: utf-8 -*-
# Copyright (c) 2025-2025 Huawei Technologies Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import stat
import argparse

from .constants import EXT_SIZE_MAPPING, TENSOR_MAX_SIZE
from .utils import confirmation_interaction


def _has_soft_link(path: str) -> bool:
    abs_path = os.path.abspath(path)
    norm_path = os.path.normpath(abs_path)
    
    current = os.path.sep
    for part in norm_path.split(os.sep)[1:]:
        if not part:
            continue
        current = os.path.join(current, part)
        if os.path.islink(current):
            return True
    
    return os.path.islink(norm_path)


def _check_file_size_based_on_ext(path):
    ext = os.path.splitext(path)[1]
    size = os.path.getsize(path)

    if ext in EXT_SIZE_MAPPING:
        if size > EXT_SIZE_MAPPING[ext]:
            return False
    else:
        if size > TENSOR_MAX_SIZE:
            confirmation_prompt = "The file %r is larger than expected. " \
                                "Attempting to read such a file could potentially impact system performance.\n" \
                                "Please confirm your awareness of the risks associated with this action (y/n): " % path
            return confirmation_interaction(confirmation_prompt)

    return True


def _normal_user_extra_checks(
        path: str, 
        st: os.stat_result, 
        is_dir: True, 
        require_executable: True,
        error_type: Exception
    ):
    if _has_soft_link(path):
        raise error_type(f"Path contains soft links: {path!r}")
    
    access_mode = os.X_OK if require_executable else os.R_OK
    error_msg = "Path not executable" if require_executable else "Path not readable"

    if not os.access(path, access_mode):
        raise error_type(f"{error_msg}: {path!r}")
    
    if st.st_mode & (stat.S_IWGRP | stat.S_IWOTH):
        raise error_type(f"Insecure group or others write permissions: {path!r}")
    
    if st.st_uid not in (0, os.geteuid()):
        raise error_type(f"Path ownership mismatch: {path!r}")
    
    if not is_dir and not _check_file_size_based_on_ext(path):
        raise error_type(f"File too large: {path!r}")
    
    return os.path.realpath(path)


def _common_security_checks(
    path: str,
    *,
    is_dir: bool,
    require_executable: bool,
    raise_argparse: bool = False
) -> str:
    if not isinstance(path, str):
        raise TypeError("Expected path to be 'str', got %r instead" % type(path))
    
    error_type = argparse.ArgumentTypeError if raise_argparse else OSError
    
    try:
        st = os.stat(path)
    except OSError as e:
        raise error_type(f"File not found or no permission: {path!r}") from e
    
    file_type_is_valid = stat.S_ISDIR(st.st_mode) if is_dir else stat.S_ISREG(st.st_mode)
    error_msg = "Not a directory" if is_dir else "Not a regular file"
    
    if not file_type_is_valid:
        raise error_type(f"{error_msg}: {path!r}")
    
    if os.geteuid() == 0:
        return os.path.realpath(path)

    return _normal_user_extra_checks(path, st, is_dir, require_executable, error_type)


def read_file_common_check(path: str, *, raise_argparse: bool = True):
    return _common_security_checks(
        path, 
        is_dir=False, 
        require_executable=False, 
        raise_argparse=raise_argparse
    )


def execute_file_common_check(path: str, *, raise_argparse: bool = True):
    return _common_security_checks(
        path, 
        is_dir=False, 
        require_executable=True, 
        raise_argparse=raise_argparse
    )
    

def list_dir_common_check(path: str, *, raise_argparse: bool = True):
    return _common_security_checks(
        path, 
        is_dir=True, 
        require_executable=False, 
        raise_argparse=raise_argparse
    )


def traverse_dir_common_check(path: str, *, raise_argparse: bool = True):
    return _common_security_checks(
        path, 
        is_dir=True, 
        require_executable=True, 
        raise_argparse=raise_argparse
    )
